/* vim: set sw=8 ts=8 si et: */
/* $Revision: 1.3 $, last changed: $Date: 1998/06/02 18:06:53 $ File: $RCSfile: main.c,v $ */

#include <stdio.h>
#include <unistd.h>
#include <string.h>
#include <strings.h> /*solaris needs this*/
#include <stdlib.h>
#include "common.h"
#include "bin_asc.h"
/*for the cbc init. vector:*/
#include <time.h>
static int opt_q=0;


static int gethexpw(char *result_bin_key,char *ascii_hex_str,int maxresultbuflen);

void help(char *myfilename)
{
        char *tmp;
        tmp=rindex(myfilename,'/');
        if (tmp == NULL){
                tmp=myfilename;
        }else{
                /*point to charcater after the slash*/
                tmp++;
        }
        printf("%s -- encrypt or decrypt files
USAGE: %s [-hbdeq] [-p passw] [-x hexpassw] [infile] [outfile]        

As filter:   
               %s [-hbdeq] [-p passw] [-x hexpassw]
Read file, write to stdout: 
               %s [-hbdeq] [-p passw] [-x hexpassw] infile
Read file, write file: 
               %s [-hbdeq] [-p passw] [-x hexpassw] infile outfile

OPTIONS: -h this help message
         -m use ECB mode (CBC with time stamp init vector is default)
         -b output binary data on encryption. 
            The default is a base 64 format.
         -e encrypt
         -d decrypt (default)
         -p specify password as plain text 
         -q quiet, do not print what the program is currently doing.
         -x specify password as hex number
\n",tmp,tmp,tmp,tmp,tmp);
#ifdef VERINFO
        puts(VERINFO); 
#endif  
        exit(0);

}

int main(int argc, char *argv[])
{
        int i;
        UBYTE_08bits workbuffer[8];
        UBYTE_08bits workbuffer_d[8];
        UBYTE_08bits chainbuffer[8];
        UBYTE_08bits chainbuffer_d[8];
        struct tm *timestrptr;
        time_t tme;
        int bytesread;
        int cbc=1; /* chaining is default*/
        int asc=1; /* ascii base 64 output is default */
        int decrypt=1; /* decrypt is default */
        FILE *instream,*outstream;
        char *infile,*outfile;
        int keylen=0;
        char *keystring = NULL;
        /* max 448 bit key, we need to +1 as getpass adds a \0 : */
        char key[MAXKEYBYTES+1]="";

        /* The following things are used for getopt: */
        extern char *optarg;
        extern int optind;
        extern int opterr;
        int c;


        opterr = 0;
        while ((c = getopt(argc, argv, "hbedmp:qx:")) != -1) {
                switch (c) {
                case 'b':
                        /*output plain binary*/
                        asc=0;
                        break;
                case 'm':
                        /*work on blocks and do not chain, ecb*/
                        cbc=0;
                        break;
                case 'e':
                        /*encrypt*/
                        decrypt=0;
                        break;
                case 'd':
                        /*decrypt is default*/
                        break;
                case 'p':
                        /*get the password*/
                        strncpy(key,optarg,sizeof(key));
                        keystring=key;
                        keylen=strlen(keystring);
                        break;
                case 'q':
                        /*quiet*/
                        opt_q=1;
                        break;
                case 'x':
                        /*get the hex password*/
                        keylen=gethexpw(key,optarg,MAXKEYBYTES+1);
                        keystring=key;
                        break;
                case 'h':
                        help(argv[0]);
                        /*no break, help does not return */
                case '?':
                        fprintf(stderr, "ERROR: No such option. -h for help.\n");
                        exit(1);
                        /*no default action for case */
                }
        }
        /* max 2 cmd line argumnts: */
        if (argc - optind > 2) help(argv[0]);

        if (opt_q==0){
                if (decrypt){
                        fprintf(stderr,"decrypt...\n");
                }else{
                        fprintf(stderr,"encrypt..\n");
                }
        }
        if (keystring == NULL){
                /*no password was given on cmd line*/
                keystring=mygetpass("Enter key: ");
                strncpy(key,keystring,sizeof(key));
                keystring=mygetpass("Enter key again: ");
                if (strncmp(keystring,key,sizeof(key))!= 0){
                        fprintf(stderr,"Keys differ, quit...\n");
                        exit(1);
                }
                keylen=strlen(keystring);
        }
        Initializekey(key,keylen);
        /* Blot out original key */
        memset(key,0,MAXKEYBYTES);
        memset(keystring,0,MAXKEYBYTES);
        /*
         *initialize chain buffer
         */
        memset(chainbuffer,0,sizeof(chainbuffer));
        /* now the CBC initialisation vector*/
        if (cbc){
                tme=time(NULL);
                timestrptr=gmtime(&tme);
                /*exactly 8 bytes, including \0 */
                sprintf(chainbuffer,"%03d%02d%02d",timestrptr->tm_yday,timestrptr->tm_min,timestrptr->tm_sec);
        }

        /*defaults:*/
        instream=stdin;
        outstream=stdout;
        if (1 == argc - optind) {
                /*read infile and write stdout */
                infile=argv[optind];
                instream=fopen(infile,"r");
                if (instream==NULL){
                        fprintf(stderr,"ERROR: Could not open file %s for reading\n",infile);
                        exit(1);
                }
        } else if (2 == argc - optind) {
                /*read infile and write outfile */
                infile=argv[optind];
                instream=fopen(infile,"r");
                if (instream==NULL){
                        fprintf(stderr,"ERROR: Could not open file %s for reading\n",infile);
                        exit(1);
                }

                outfile=argv[optind+1];
                outstream=fopen(outfile,"w");
                if (outstream==NULL){
                        fprintf(stderr,"ERROR: Could not open file %s for writing\n",outfile);
                        exit(1);
                }
        } 
        
        if (decrypt){
                /*the first block is the chain buffer:*/
                if (cbc){
                        if (asc){
                                open64base_r(instream);
                                bytesread = read64base(chainbuffer,instream);
                        }else{
                                bytesread = fread(chainbuffer,1,8,instream);
                        }
                }
                if (asc){
                        if (!cbc) open64base_r(instream);
                        bytesread = read64base(workbuffer,instream);
                }else{
                        bytesread = fread(workbuffer,1,8,instream);
                }
                while(1){
                        if(cbc){ /*save incomming ciphertext for chain*/
                                memcpy(chainbuffer_d,workbuffer,8);
                        }
                        do_decipher(workbuffer);
                        if(cbc){
                                /* CBC mode; chain in last cipher word */
                                for(i=0; i<8; i++){
                                        workbuffer[i] ^=chainbuffer[i];
                                }
                                memcpy(chainbuffer,chainbuffer_d,8);
                        }
                        memcpy(workbuffer_d,workbuffer,8);
                        /*the file can only have a multiple of 8 bytes
                         *if the next read fails then we must only write
                         *this block or parts of it.
                         */
                        if (asc){
                                bytesread = read64base(workbuffer,instream);
                        }else{
                                bytesread = fread(workbuffer,1,8,instream);
                        }
                        if(bytesread != 8){   /* Can "only" be 0 if not 8 */
                                if(bytesread < 0 || bytesread > 7){
                                        fprintf(stderr,"Corrupted file or wrong key\n");
                                        exit(1);
                                } else{
                                        /*get the real length from last byte*/
                                        bytesread = workbuffer_d[7];
                                        if(bytesread != 0){
                                                fwrite(workbuffer_d,1,bytesread,outstream);
                                        }
                                }
                                exit(0); /*unix will close files*/
                        }
                        /* Now okay to write previous buffer */
                        fwrite(workbuffer_d,1,8,outstream);
                }
        }else{
                if (asc) open64base_w(outstream);
                /*encrypt:*/
                if(cbc){
                        /*the first 8 bytes of cipher text hold 
                         *the CBC initialisation vector*/
                        do_encipher(chainbuffer);
                        if (asc){
                                write64base(chainbuffer,outstream);
                        }else{
                                if (fwrite(chainbuffer,1,8,outstream)!=8){
                                        fprintf(stderr,"ERROR: unable to write output\n");
                                        exit(1);
                                }
                        }
                }
                while(1){
                        if((bytesread = fread(workbuffer,1,8,instream)) != 8){
                                /*we are at end of file, put residual byte count
                                 *in the last block. Note: garbage is left in the 
                                 *other bytes. This is ok as it is stripped out 
                                 *at decrypt time.
                                 */
                                 workbuffer[7] = bytesread;
                        }
                        if(cbc){
                                /* CBC mode; chain in last cipher word */
                                for(i=0; i<8; i++){
                                        workbuffer[i] ^=chainbuffer[i];
                                }
                        }
                        /*byteswap and endiness must be adjusted
                         *in the various algorithms as some might
                         *work on 16 bit numbers and others on 
                         *32 bit numbers*/
                        do_encipher(workbuffer);
                        if(cbc){ /*save outgoing ciphertext for chain*/
                                memcpy(chainbuffer,workbuffer,8);
                        }
                        if (asc){
                                write64base(workbuffer,outstream);
                        }else{
                                if (fwrite(workbuffer,1,8,outstream)!=8){
                                        fprintf(stderr,"ERROR: unable to write output\n");
                                        exit(1);
                                }
                        }
                        if (bytesread!=8){
                                /*we are at end of file and done*/
                                if (asc) close64base_w(outstream);
                                exit(0); /*unix will close files*/
                        }
                }
                if (asc) close64base_w(outstream);
        }
        cipher_cleanup();

        return (0);
}
/*
 * Convert hex/ascii nybble (4bit) to binary 
 */
static int htoa(char c)
{
        if(c >= '0' && c <= '9')
                return c - '0';
        if(c >= 'a' && c <= 'f')
                return 10 + c - 'a';
        if(c >= 'A' && c <= 'F')
                return 10 + c - 'A';
        fprintf(stderr,"ERROR: non hex \"%c\", only 0-9a-f allowed\n",c);
        exit(1);
}
/*
 *Convert a ascii string representaion of a longer hex number 
 *into a binary byte stream
 *Returns the length of result_bin_key
 *maxresultbuflen is the size in bytes of result_bin_key (memory must 
 *be allocated already)
 *Parts of the key are discared if necessary to make the result fit into
 *result_bin_key
 */
int gethexpw(char *result_bin_key,char *ascii_hex_str,int maxresultbuflen)
{
        int len,rlen;
        len = strlen(ascii_hex_str);
        if (len < 6){
                fprintf(stderr,"ERROR: string too short\n");
                exit(1);
        }
        rlen=0;
        while(len){
                if (maxresultbuflen <= rlen){
                        return(rlen);
                }
                *result_bin_key = htoa(*ascii_hex_str)  << 4;
                ascii_hex_str++;
                len--;
                /* we have may have an uneven length */
                if(*ascii_hex_str){
                         *result_bin_key |=htoa(*ascii_hex_str);
                         ascii_hex_str++;
                         len--;
                }
                result_bin_key++;
                rlen++;
        }
        return(rlen);
}         

